/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>
#include "kiss_fft.h"
#include "chroma.h"

chroma::chroma( double set_fmin, double set_fmax, int set_bpo, double set_Fs ) {

	/* Declarations */	
	int n,k;
	double thresh,Q,amp;
	int ilength,start;

	/* Set parameters */
	fmin = set_fmin;
	fmax = set_fmax;
	bpo = set_bpo;
	Fs = set_Fs;

	/* Calculate other parameters */
	thresh = 0.0054;
	Q = 1.0/(pow(2,1.0/(double)bpo)-1.0);
	K = (int)ceil((double)bpo*log2(fmax/fmin));
	fftlength = (int)pow(2,ceil(log2(ceil(Q*(double)Fs/(double)fmin))));

	/* Init fft */
	infft = new kiss_fft_cpx[fftlength];
	outfft = new kiss_fft_cpx[fftlength];
	cfg = kiss_fft_alloc( fftlength , 0, 0, 0 );

	/* Init kernel */
	kernel[0] = new double[fftlength*K];
	kernel[1] = new double[fftlength*K];
	for (n=0;n<fftlength*K;n++) {
		kernel[0][n] = 0.0;
		kernel[1][n] = 0.0;
	}
	for (n=0;n<fftlength;n++) {
		infft[n].r = 0.0;
		infft[n].i = 0.0;
	}
	kernelfirst = new double[K];
	kernellast = new double[K];

	/* Compute kernel */
	for (k=K-1;k>=0;k--) {

		ilength = (int)ceil( Q*Fs/(fmin*pow(2.0,(double)k/(double)bpo)));
		if (ilength%2==0)
			start = (fftlength/2) - (ilength/2) - 1;
		else
			start = (fftlength/2) - ((ilength+1)/2) - 1;
	
		for (n=0;n<ilength;n++) {
			infft[start+n].r = ((0.54-0.46*cos(2.0*M_PI*(double)n/(double)(ilength-1)))/(double)ilength) * cos( 2.0*M_PI*Q*(double)n/(double)ilength);
			infft[start+n].i = ((0.54-0.46*cos(2.0*M_PI*(double)n/(double)(ilength-1)))/(double)ilength) * sin( 2.0*M_PI*Q*(double)n/(double)ilength);
		}

		kiss_fft( cfg , infft , outfft );

		for (n=0;n<ilength;n++) {
			amp = sqrt( outfft[n].r*outfft[n].r + outfft[n].i*outfft[n].i );
			if (amp<=thresh) {
				kernel[0][k*fftlength+n] = 0.0;
				kernel[1][k*fftlength+n] = 0.0;
			} else {
				kernel[0][k*fftlength+n] = outfft[n].r / (double)fftlength;
				kernel[1][k*fftlength+n]  = - outfft[n].i / (double)fftlength;
			}
		}

		kernelfirst[k] = 0;
		kernellast[k] = fftlength-1;
		for (n=0;n<fftlength;n++) {
			if (kernel[0][k*fftlength+n]!=0||kernel[1][k*fftlength+n]!=0) {
				kernelfirst[k] = n;
				break;
			}
		}
		for (n=fftlength-1;n>=0;n--) {
			if (kernel[0][k*fftlength+n]!=0||kernel[1][k*fftlength+n]!=0) {
				kernellast[k] = n;
				break;
			}
		}
	}

	/* Init window */
	window = new double[fftlength];
	for (n=0;n<fftlength;n++) window[n] = 0.5*(1.0-cos(2.0*M_PI*(double)(n+1)/(double)(fftlength+1)));

	/* Init constant Q output */
	cqout = new double[K];
}

chroma::~chroma( void ) {

	/* Clean memory */
	free( cfg );
	delete[]( infft );
	delete[]( outfft );
	delete[]( kernel[0] );
	delete[]( kernel[1] );
	delete[]( kernelfirst );
	delete[]( kernellast );		
	delete[]( window );
	delete[]( cqout );
}

void chroma::compute( double *in, double *out ) {

	/* Declarations */
	int n,k;
	double tempRe;
	double tempIm;
	int numoctaves;

	/* Window  */
	for (n=0;n<fftlength;n++) {
		infft[n].r = in[n] * window[n];
		infft[n].i = 0.0;
	}

	/* FFT */
	kiss_fft( cfg , infft , outfft );

	/* Constant Q */
	for (k=0;k<K;k++) {
		tempRe = 0;
		tempIm = 0;
		for (n=kernelfirst[k];n<=kernellast[k];n++) {
			tempRe += outfft[n].r*kernel[0][k*fftlength+n]-outfft[n].i*kernel[1][k*fftlength+n];
			tempIm += outfft[n].i*kernel[0][k*fftlength+n]+outfft[n].r*kernel[1][k*fftlength+n];
		}
		cqout[k] = sqrt(tempRe*tempRe+tempIm*tempIm);
	}

	/* Chromagram */
	numoctaves = (int)floor((double)K/(double)bpo);
	for (n=0;n<bpo;n++) out[n] = 0.0;
	for (k=0;k<numoctaves;k++)
		for (n=0;n<bpo;n++)
			out[n] += cqout[k*bpo+n];
}

