/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>
#include "kiss_fft.h"
#include "mfcc.h"

mfcc::mfcc( double set_fs, int set_winlen, double *set_window, double set_lowfreq, double set_highfreq, int set_numfilters, int set_numcoefs ) {

	/* Declarations */
	int n,m;
	double *freqs,*filtersfreq,*lower,*center,*upper,*triangle_heights;
	double lowfreqmel,highfreqmel;

	/* Set parameters */
	fs = set_fs;
	winlen = set_winlen;
	winlen2 = winlen >> 1;
	lowfreq = set_lowfreq;
	highfreq = set_highfreq;
	numfilters = set_numfilters;
	numcoefs = set_numcoefs;

	/* Init window */
	window = new double[winlen];
	for (n=0;n<winlen;n++) window[n] = set_window[n];

	/* Init arrays */
	amp = new double[winlen2];
	filterbank = new double[numfilters*winlen2];
	fbout = new double[numfilters];
	dct = new double[numfilters*numfilters];

	/* Init fft */
	infft = new kiss_fft_cpx[winlen];
	outfft = new kiss_fft_cpx[winlen];
	cfg = kiss_fft_alloc( winlen , 0, 0, 0 );

	/* Init temporary arrays */
	freqs = new double[winlen2];
	filtersfreq = new double[numfilters+2];
	lower = new double[numfilters];
	center = new double[numfilters];
	upper = new double[numfilters];
	triangle_heights = new double[numfilters];

	/* Init filterbank */
	for (n=0;n<winlen2;n++) freqs[n] = (double)n*fs/(double)winlen;
	lowfreqmel = log(1.0+lowfreq/700.0)*1127.01048;
	highfreqmel = log(1.0+highfreq/700.0)*1127.01048;
	for (n=0;n<numfilters+2;n++)
		filtersfreq[n] = 700.0*( exp( (lowfreqmel+((double)n/(double)(numfilters+1))*(highfreqmel-lowfreqmel)) / 1127.01048 ) - 1 );
	for (n=0;n<numfilters;n++) {
		lower[n] = filtersfreq[n];
		center[n] = filtersfreq[n+1];
		upper[n] = filtersfreq[n+2];
		triangle_heights[n] = 2/(upper[n]-lower[n]);
	}
	for ( n = 0; n < numfilters ; n++) {
		for ( m=0; m< winlen2 ; m++) {
			filterbank[n*winlen2+m] = 0.0;
			if ((freqs[m] > lower[n])&& (freqs[m] <= center[n])) {
				filterbank[n*winlen2+m] = triangle_heights[n] * ((freqs[m] - lower[n])/(center[n] - lower[n]));
			}
			if ((freqs[m] > center[n]) && (freqs[m] <= upper[n])) {
				filterbank[n*winlen2+m] = triangle_heights[n] * ((upper[n] - freqs[m])/(upper[n] - center[n]));
			}
		}
	}

	/* Delete temporary arrays */
	delete[]( freqs );
	delete[]( filtersfreq );
	delete[]( lower );
	delete[]( center );
	delete[]( upper );
	delete[]( triangle_heights );

	/* Init DCT */
	for ( n=0 ; n<numfilters ; n++ ) {
		for ( m=0 ; m<numfilters ; m++ ) {
			dct[n*numfilters+m] = (1/sqrt(numfilters/2)) * cos( n * (2*m+1) * M_PI/2/numfilters );
			if (n==0) dct[n*numfilters+m] *= sqrt(2)/2;
		}
	}
}

mfcc::~mfcc( void ) {

	/* Clean memory */
	free( cfg );
	delete[]( infft );
	delete[]( outfft );
	delete[]( window );
	delete[]( filterbank );
	delete[]( dct );
	delete[]( amp );
	delete[]( fbout );
}

void mfcc::compute( double *in, double *out ) {

	/* Declarations */
	int n,m;
	double eps = 1e-16;

	/* Window  */
	for (n=0;n<winlen;n++) {
		infft[n].r = in[n] * window[n];
		infft[n].i = 0.0;
	}

	/* FFT */
	kiss_fft( cfg , infft , outfft );

	/* Amplitude */
	for (n=0;n<winlen2;n++) {
		amp[n] = sqrt( outfft[n].r*outfft[n].r + outfft[n].i*outfft[n].i );
	}

	/* Filterank */
	for ( n=0 ; n<numfilters ; n++ ) {
		fbout[n] = 0.0;
		for ( m=0 ; m<winlen2 ; m++ )
			fbout[n] += filterbank[n*winlen2+m] * amp[m];
	}

	/* Logarithm */
	for ( n=0 ; n<numfilters ; n++ )
		fbout[n] = log10( fbout[n] + eps );

	/* DCT */
	for ( n=0 ; n<numcoefs ; n++ ) {
		out[n] = 0.0;
		for ( m=0 ; m<numfilters ; m++ )
			out[n] += dct[n*numfilters+m] * fbout[m];
	}
}

